from pathlib import Path
from dataclasses import dataclass, field
import logging


def default_image_transforms(max_size=5000) -> dict:
    return {
        '': {
            'max_size': (max_size, max_size),
        },
    }


@dataclass
class UrsusConfig():
    content_path: Path = Path('content').resolve()
    templates_path: Path = Path('templates').resolve()
    output_path: Path = Path('output').resolve()

    # The URL of this website's root, without a trailing slash. For example, https://allaboutberlin.com
    site_url: str = ''

    # The URL extension of HTML pages. Change this if your server changes or removes the file extension.
    html_url_extension: str = '.html'

    # The base URL prepended to all markdown [[wikilinks]], without a trailing slash.
    # For example, https://allaboutberlin.com/glossary
    wikilinks_base_url: str = '/'
    wikilinks_url_suffix: str = ''
    wikilinks_url_builder: callable = None
    wikilinks_html_class: str = None

    # Minify Javascript and CSS
    minify_js: bool = False
    minify_css: bool = False

    # Rebuilds the output faster by only rebuilding templates for the changed files.
    # Related pages (like index pages) will not be rebuild, even though they could change.
    # If false, the pages that definitely changed are still rebuilt before others.
    fast_rebuilds: bool = False

    # Transforms applied to your content images
    image_transforms: dict = field(default_factory=default_image_transforms)

    """
    Parametres to generate a search index for lunr.js
    Example:
    {
        'indexed_fields': ('body', ),  # Index these fields only
        'indexes': [
            {
                'uri_pattern': '*.md',  # Index entries with URIs that match this glob pattern
                'returned_fields': ('body', 'url', ),  # Return these fields only in the document list
                'boost': 1,  # Documents matching this pattern should be boosted n times over others
            },
        ]
    }
    """
    lunr_indexes: dict = field(default_factory=dict)
    lunr_index_output_path: Path = Path('search-index.json')  # Relative to output_path

    generator: str = 'ursus.generators.static.StaticSiteGenerator'

    # The processors that update the context with extra data
    context_processors: tuple = (
        'ursus.context_processors.stale.StaleEntriesProcessor',
        'ursus.context_processors.markdown.MarkdownProcessor',
        'ursus.context_processors.get_entries.GetEntriesProcessor',
        'ursus.context_processors.related.RelatedEntriesProcessor',
        # 'ursus.context_processors.git_date.GitDateProcessor',
    )
    context_globals: dict = field(default_factory=dict)

    # The renderers that take your templates and content, and populate the output dir
    renderers: tuple = (
        'ursus.renderers.static.StaticAssetRenderer',
        'ursus.renderers.image.ImageTransformRenderer',
        'ursus.renderers.jinja.JinjaRenderer',
        'ursus.renderers.lunr.LunrIndexRenderer',
    )

    # Linters look for errors in your content
    linters: tuple = (
        'ursus.linters.markdown.MarkdownLinksLinter',
    )

    # Filter functions available in Jinja templates. The key is the filter name, and the value is a function
    jinja_filters = {}

    logging = {
        'datefmt': '%Y-%m-%d %H:%M:%S',
        'fmt': '%(asctime)s %(levelname)s [%(name)s:%(lineno)d] %(message)s',
        'level': logging.INFO,
    }


config = UrsusConfig()



from functools import partial
from pathlib import Path
from ursus.config import config
import logging
import re


logger = logging.getLogger(__name__)


class Linter():
    def lint(self, file_path: Path, fix_errors: bool = False):
        """
        Lints the content for errors.
        """
        raise NotImplementedError

    def log_error(self, file_path: int, line_no: int, message: str, level=logging.WARNING):
        color = {
            logging.DEBUG: '',
            logging.INFO: '\033[37m\033[0;100m',
            logging.WARNING: '\033[1;90m\033[43m',
            logging.ERROR: '\033[37m\033[41m',
            logging.CRITICAL: '\033[37m\033[41m',
        }[level]
        logger.log(level, f"    {color}:{line_no}\033[0m {message}")

    def log_substitution(self, file_path: int, line_no: int, old: str, new: str):
        logger.info(f"      \033[0;31m- {old}\033[0m")
        logger.info(f"      \033[0;32m+ {new}\033[0m")


class RegexLinter(Linter):
    file_suffixes = None
    regex = re.compile(r'')

    def lint(self, file_path: Path, fix_errors: bool = False):
        if not self.file_suffixes or file_path.suffix in self.file_suffixes:
            logger.info(f"\033[1mLinting {str(file_path)}\033[0m")
            with (config.content_path / file_path).open() as file:
                old_lines = file.readlines()

            new_lines = []
            for line_no, line in enumerate(old_lines):
                sub_func = partial(self.handle_match, file_path, line_no, fix_errors)
                new_lines.append(self.regex.sub(sub_func, line))

            with (config.content_path / file_path).open('w') as file:
                file.writelines(new_lines)

    def handle_match(self, file_path: int, line_no: int, fix_errors: bool, match: re.Match):
        raise NotImplementedError



from ursus.utils import get_files_in_path
from ursus.config import config


class ContextProcessor:
    def process(self, context: dict, changed_files: set = None) -> dict:
        return context


class EntryContextProcessor(ContextProcessor):
    def process(self, context: dict, changed_files: set = None) -> dict:
        for file_path in get_files_in_path(config.content_path, changed_files):
            entry_uri = str(file_path)
            entry = context['entries'].get(entry_uri, {})
            context['entries'][entry_uri] = self.process_entry(entry_uri, entry)

            assert context['entries'][entry_uri] is not None, \
                "{type(self).__name__}.process_entry is returning None instead of the entry context."

        return context

    def process_entry(self, entry_uri: str, entry_context: dict) -> dict:
        raise NotImplementedError



from . import Renderer
from PIL import Image
from ursus.config import config
from ursus.utils import get_files_in_path, make_image_thumbnail, make_pdf_thumbnail, is_image, is_pdf, is_svg, \
    copy_file, get_image_transforms
import logging


logger = logging.getLogger(__name__)


class ImageTransformRenderer(Renderer):
    """
    Resizes images and generate PDF thumbnails
    """
    def __init__(self):
        super().__init__()

    def get_files_to_transform(self):
        return [
            f for f in get_files_in_path(config.content_path)
            if is_image(config.content_path / f) or is_pdf(config.content_path / f)
        ]

    def render(self, context: dict, changed_files: set = None) -> set:
        logger.info("Rendering image transforms...")

        files_to_keep = set()

        for input_path in self.get_files_to_transform():
            abs_input_path = config.content_path / input_path
            has_changed = changed_files is None or abs_input_path in changed_files

            for transform in get_image_transforms(input_path):
                output_path = transform['output_path']
                abs_output_path = config.output_path / output_path
                max_size = transform['max_size']

                if has_changed and not abs_output_path.exists():
                    if is_pdf(abs_input_path):
                        if abs_output_path.suffix.lower() == '.pdf':
                            logger.info('Copying %s to %s', str(input_path), str(output_path))
                            copy_file(abs_input_path, abs_output_path)
                        else:
                            logger.info('Generating %s preview as %s', str(input_path), str(output_path))
                            make_pdf_thumbnail(config.content_path / input_path, max_size, abs_output_path)
                    elif is_svg(abs_input_path):
                        if abs_output_path.suffix.lower() == '.svg':
                            logger.info('Copying %s to %s', str(input_path), str(output_path))
                            copy_file(abs_input_path, abs_output_path)
                        else:
                            logger.warning(f"Can't convert {str(input_path)} to {abs_output_path.suffix}. Ignoring.")
                            continue
                    else:
                        logger.info('Converting %s to %s', str(input_path), str(output_path))
                        with Image.open(abs_input_path) as pil_image:
                            make_image_thumbnail(pil_image, max_size, abs_output_path)

                files_to_keep.add(output_path)

        return files_to_keep



from . import Renderer
from ursus.config import config
from ursus.utils import get_files_in_path, copy_file
import logging


logger = logging.getLogger(__name__)


class StaticAssetRenderer(Renderer):
    """
    Copies static assets in `templates_path` to `output_path`.
    """
    ignored_suffixes = ('.jinja', )

    def get_assets_to_copy(self, changed_files: set = None):
        return [
            f for f in get_files_in_path(config.templates_path)
            if f.suffix not in self.ignored_suffixes
        ]

    def render(self, context: dict, changed_files: set = None) -> set:
        files_to_keep = set()
        for asset_path in self.get_assets_to_copy():
            abs_output_path = config.output_path / asset_path

            if changed_files is None or config.templates_path / asset_path in changed_files:
                logger.info('Copying asset %s', str(asset_path))
                copy_file(config.templates_path / asset_path, abs_output_path)
            files_to_keep.add(asset_path)

        return files_to_keep



from requests.exceptions import ConnectionError
from ursus.config import config
from ursus.linters import RegexLinter
import logging
import re
import requests


class MarkdownLinksLinter(RegexLinter):
    file_suffixes = ('.md',)

    # Matches [], supports escaped brackets, ignores ![] images.
    first_half = r"(?P<first_half>!?\[(?P<text>([^\]]|\\\])*)(?<!\\)\])"

    # Matches (), supports escaped parentheses
    second_half = r"(?P<second_half>\((?P<url_group>([^\)]|\\\))*)(?<!\\)\))"

    regex = re.compile(first_half + second_half)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.response_cache = {}

    def escape_url(self, url: str) -> str:
        return url.replace('(', '\\(').replace(')', '\\)')

    def unescape_url(self, url: str) -> str:
        return url.replace('\\(', '(').replace('\\)', ')')

    def handle_match(self, file: int, line: int, fix_errors: bool, match: re.Match):
        original_text = match.group(0)
        new_text = original_text
        text = match['text'].strip()
        url = None
        title = None
        is_image = match['first_half'].startswith('!')

        if match['url_group']:
            parts = match['url_group'].split(" ", maxsplit=1)
            url = parts[0]
            if len(parts) == 2:
                title = parts[1]

        # Validate the link text
        if text in ('click', 'here', 'this', 'link'):
            self.log_error(file, line, f"Link text is not informative: {text}", logging.INFO)
        elif not text:
            error = f"Image has no alt text: {original_text}" if is_image else f"Link text is empty: {original_text}"
            self.log_error(file, line, error, logging.ERROR)

        # Validate the title
        if title is not None:
            if not (title.startswith('"') and title.endswith('"')):
                self.log_error(file, line, f"Title is not quoted: {original_text}", logging.ERROR)
            elif title == title.strip('"'):
                self.log_error(file, line, f"Title is empty: {original_text}", logging.ERROR)

        # Validate the URL
        if not url:
            self.log_error(file, line, f"Missing URL: {original_text}", logging.ERROR)
        elif not url.startswith(('/', '#', 'http://', 'https://', 'mailto:', 'tel:')):
            self.log_error(file, line, f"Relative or invalid URL: {original_text}", logging.WARNING)
        else:
            if url.startswith(('http://', 'https://')):
                try:
                    cleaned_url = self.unescape_url(url)
                    if cleaned_url not in self.response_cache:
                        self.response_cache[cleaned_url] = requests.get(cleaned_url, timeout=5, headers={
                            'User-Agent': "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/110.0.0.0 Safari/537.36"
                        })
                    response = self.response_cache[cleaned_url]
                    status_code = response.status_code
                except ConnectionError:
                    self.log_error(file, line, f"Connection error: {cleaned_url}", logging.ERROR)
                except requests.exceptions.RequestException as exc:
                    self.log_error(file, line, f"URL {type(exc).__name__}: {cleaned_url}", logging.ERROR)
                else:
                    if status_code in (404, 410):
                        self.log_error(file, line, f"URL returns HTTP {status_code}: {cleaned_url}", logging.ERROR)
                        archived_versions = requests.get(
                            'https://archive.org/wayback/available', params={'url': cleaned_url}
                        ).json()
                        latest_archive = archived_versions.get('archived_snapshots', {}).get('closest')
                        if latest_archive and latest_archive['status_code'] == '200':
                            new_text = match['first_half'] + match['second_half'].replace(
                                url, self.escape_url(latest_archive['url']))
                    elif status_code >= 400:
                        severity = logging.WARNING if status_code in (403, 503) else logging.ERROR
                        self.log_error(file, line, f"URL returns HTTP {status_code}: {cleaned_url}", severity)
                    elif response.history and response.history[-1].status_code == 301:
                        self.log_error(file, line, f"URL redirects: {cleaned_url} -> {response.url}", logging.INFO)
                        new_text = match['first_half'] + match['second_half'].replace(url, self.escape_url(response.url))
            elif url.startswith('/'):
                if is_image and not (config.content_path / url.lstrip('/')).exists():
                    self.log_error(file, line, f"Image not found: {url}", logging.ERROR)
                else:
                    pass
            elif url.startswith('#'):
                pass
            elif url.startswith(('mailto:', 'tel:')):
                pass
            else:
                self.log_error(file, line, f"Relative or invalid URL: {url}", logging.WARNING)

        if new_text != original_text:
            self.log_substitution(file, line, original_text, new_text)
        return new_text



from . import ContextProcessor
from operator import itemgetter


class GetEntriesProcessor(ContextProcessor):
    """
    Adds the get_entries() method to the context root
    """
    def process(self, context: dict, changed_files: set = None) -> dict:
        def get_entries(namespace, sort_by=None, reverse=False):
            entries = context['entries'].items()
            if namespace:
                entries = {
                    uri: value for uri, value in entries
                    if uri.startswith(namespace + '/')
                }

            entries = entries.values()
            if sort_by:
                if callable(sort_by):
                    sorter = sort_by
                elif isinstance(sort_by, str):
                    sorter = itemgetter(sort_by)
                else:
                    sorter = lambda x: list(filter(None, [x.get(key) for key in sort_by]))[0]
                entries = sorted(entries, key=sorter, reverse=reverse)

            return entries

        if get_entries not in context:
            context['get_entries'] = get_entries

        return context



from . import Renderer
from jinja2 import Environment, FileSystemLoader, nodes, pass_context, select_autoescape, StrictUndefined
from jinja2.exceptions import TemplateSyntaxError
from jinja2.ext import Extension, do
from markupsafe import Markup
from ordered_set import OrderedSet
from pathlib import Path
from rjsmin import jsmin
from rcssmin import cssmin
from ursus.config import config
from ursus.utils import get_files_in_path, make_picture_element, is_ignored_file
from xml.etree import ElementTree
import logging


logger = logging.getLogger(__name__)


class JsLoaderExtension(Extension):
    """
    Jinja extension. Adds the {% js %} and {% alljs %} tags.

    All javascript code between {% js %} tags is combined and queued for future
    output. It's minified if config.minify_js is True.

    {% alljs %} outputs the queued javascript code. The code is included in
    order of addition. If the same code is queued multiple times, it's only
    output once.
    """
    tags = {"js", "alljs"}

    def __init__(self, environment):
        super().__init__(environment)
        environment.extend(js_fragments=OrderedSet())

    def parse(self, parser):
        token = next(parser.stream)

        if token.test('name:alljs'):
            call = self.call_method('_render_js', args=[nodes.ContextReference()])
            return nodes.Output([nodes.MarkSafe(call)]).set_lineno(token.lineno)
        else:
            body = parser.parse_statements(["name:endjs"], drop_needle=True)
            return nodes.CallBlock(
                self.call_method("_queue_js"), [], [], body
            ).set_lineno(token.lineno)

    def _render_js(self, caller):
        output = "".join(self.environment.js_fragments)
        if config.minify_js:
            output = jsmin(output)
        self.environment.js_fragments.clear()
        return Markup(output)

    def _queue_js(self, caller):
        self.environment.js_fragments.add(caller())
        return ''


class CssLoaderExtension(Extension):
    """
    Jinja extension. Adds the {% css %} tag.

    All CSS code in {% css %} tags is minified if config.minify_css is True.
    """
    tags = {"css"}

    def __init__(self, environment):
        super().__init__(environment)
        environment.extend(js_fragments=OrderedSet())

    def parse(self, parser):
        token = next(parser.stream)

        if token.test('name:css'):
            body = parser.parse_statements(["name:endcss"], drop_needle=True)
            return nodes.CallBlock(
                self.call_method("_render_css"), [], [], body
            ).set_lineno(token.lineno)

    def _render_css(self, caller):
        output = caller()
        if config.minify_css:
            output = cssmin(output)
        return Markup(output)


class ResponsiveImageExtension(Extension):
    """Jinja extension. Adds {% image 'relative/path/to/original.jpg' %} tag.

    This tag is replaced by a responsive <picture> element. The correct image transforms for the given source image are
    used.
    """
    tags = {"image"}

    def __init__(self, environment):
        super().__init__(environment)
        environment.extend(js_fragments=OrderedSet())

    def parse(self, parser):
        token = next(parser.stream)

        args = [parser.parse_expression()]

        if parser.stream.skip_if("comma"):
            args.append(parser.parse_expression())
        else:
            args.append(nodes.Const(None))

        call = self.call_method('_render_image', args)
        return nodes.Output([nodes.MarkSafe(call)]).set_lineno(token.lineno)

    @pass_context
    def _render_image(self, context, image_path, image_class):
        img_attrs = {'class': image_class} if image_class else {}
        output = make_picture_element(
            original_path=Path(image_path),
            output_path=config.output_path,
            img_attrs=img_attrs,
        )
        return Markup(ElementTree.tostring(output, encoding='unicode'))


@pass_context
def render_filter(context, value):
    return context.eval_ctx.environment.from_string(value).render(**context)


def is_entry_template(template_path: Path) -> bool:
    return template_path.with_suffix('').stem == 'entry'


def template_can_render_entry(template_path: Path, entry_uri: str) -> bool:
    return entry_uri.startswith(str(template_path.parent) + '/')


class JinjaRenderer(Renderer):
    """
    Renders all .jinja templates in the templates directory, unless their name starts with '_'.
    """
    def __init__(self):
        super().__init__()
        self.template_environment = Environment(
            loader=FileSystemLoader(config.templates_path),
            extensions=[do, JsLoaderExtension, CssLoaderExtension, ResponsiveImageExtension],
            autoescape=select_autoescape(),
            undefined=StrictUndefined
        )
        self.template_environment.filters['render'] = render_filter
        self.template_environment.filters.update(config.jinja_filters)

    def render_template(self, template_path: Path, context: dict, output_path: Path):
        """Returns an entry into a template, and saves it under output_path
        Args:
            template_path (Path): Path to the template to use, relative to output_path.
            context (dict): Context object used to render the template.
            output_path (str): Path of the generated file, relative to the output_path.
        """
        logger.info('Rendering %s', str(output_path))
        abs_output_path = config.output_path / output_path
        abs_output_path.parent.mkdir(parents=True, exist_ok=True)
        template = self.template_environment.get_template(str(template_path))
        try:
            template.stream(**context).dump(str(abs_output_path))
        except TemplateSyntaxError:
            print(context.get('entry', {}).get('body'))
            raise
        return output_path

    def get_entry_output_path(self, template_path: Path, entry_uri: str) -> Path:
        """Gets the path where the rendered template will be saved
        Args:
            template_path (Path): Template used to render the entry
            entry_uri (str): URI of the entry to render
        Returns:
            Path: Rendered template output path, relative to the output_path.
        """
        output_suffix = template_path.with_suffix('').suffix
        return Path(entry_uri).with_suffix(output_suffix)

    def render_entry(self, template_path: Path, context: dict, entry_uri: str) -> Path:
        """Returns an entry into a template, and saves it under output_path
        Args:
            template_path (Path): Path to the template to use, relative to output_path.
            context (dict): Context object used to render the template.
            entry_uri (str): URI of the entry to render
        Returns:
            Path: Path of the generated file, relative to the output_path.
        """
        specific_context = {
            **context,
            'entry': context['entries'][entry_uri]
        }
        output_path = self.get_entry_output_path(template_path, entry_uri)
        return self.render_template(template_path, specific_context, output_path)

    def render(self, context: dict, changed_files: set = None) -> set:
        template_paths = get_files_in_path(config.templates_path, suffix='.jinja')

        render_queue = OrderedSet()
        files_to_keep = set()

        changed_entry_uris = set()
        changed_templates = set()
        for file in (changed_files or set()):
            if not (file.exists() and file.is_file()):
                continue

            if file.is_relative_to(config.content_path) and not is_ignored_file(file, config.content_path):
                changed_entry_uris.add(str(file.relative_to(config.content_path)))
            elif file.is_relative_to(config.templates_path) and not is_ignored_file(file, config.templates_path):
                changed_templates.add(file.relative_to(config.templates_path))
            else:
                continue

        # Process edited entries
        for entry_uri in changed_entry_uris:
            for tp in template_paths:
                if is_entry_template(tp) and template_can_render_entry(tp, entry_uri):
                    render_queue.add(('entry', tp, entry_uri))

        # Process edited templates
        for template_path in changed_templates:
            if is_entry_template(template_path):
                for entry_uri in context['entries']:
                    if template_can_render_entry(template_path, entry_uri):
                        render_queue.add(('entry', template_path, entry_uri))
            else:
                render_queue.add(('template', template_path, template_path.with_suffix('')))

        # Process everything else
        for template_path in template_paths:
            if is_entry_template(template_path):
                for entry_uri in context['entries']:
                    if template_can_render_entry(template_path, entry_uri):
                        if config.fast_rebuilds:
                            (config.output_path / self.get_entry_output_path(template_path, entry_uri)).touch()
                        else:
                            render_queue.add(('entry', template_path, entry_uri))
            else:
                output_path = template_path.with_suffix('')  # Remove .jinja
                if config.fast_rebuilds:
                    (config.output_path / output_path).touch()
                else:
                    render_queue.add(('template', template_path, output_path))

        for render_type, template_path, value in render_queue:
            if render_type == 'entry':
                files_to_keep.add(self.render_entry(template_path, context, value))
            elif render_type == 'template':
                files_to_keep.add(self.render_template(template_path, context, value))

        return files_to_keep



from collections import UserDict
from . import ContextProcessor


class RelatedEntriesProcessor(ContextProcessor):
    """
    Entry fields that start with related_* return a list of entries, instead of
    a list of entry URIs.
    """
    class RelatedEntryReferenceDict(UserDict):
        def __init__(self, entry, all_entries):
            self.all_entries = all_entries
            super().__init__(entry)

        def __getitem__(self, key):
            if key.startswith('related_') and key in self.data:
                related_value = self.data[key]
                if isinstance(related_value, str):  # Single URI string
                    return [self.all_entries[related_value]]
                else:  # List of URI strings
                    return [self.all_entries[subvalue] for subvalue in related_value]
            return super().__getitem__(key)

    def process(self, context: dict, changed_files: set = None) -> dict:
        for uri, entry in context['entries'].items():
            if type(context['entries'][uri]) is not self.RelatedEntryReferenceDict:
                context['entries'][uri] = self.RelatedEntryReferenceDict(entry, context['entries'])

        return context



from . import EntryContextProcessor
from datetime import datetime
from markdown.extensions import Extension
from markdown.extensions.footnotes import FootnoteExtension, FN_BACKLINK_TEXT, NBSP_PLACEHOLDER
from markdown.extensions.smarty import SubstituteTextPattern
from markdown.extensions.toc import TocExtension, slugify
from markdown.extensions.wikilinks import WikiLinkExtension, build_url
from markdown.inlinepatterns import SimpleTagPattern
from markdown.postprocessors import RawHtmlPostprocessor
from markdown.preprocessors import Preprocessor
from markdown.treeprocessors import Treeprocessor, InlineProcessor
from pathlib import Path
from ursus.config import config
from ursus.utils import make_figure_element, make_picture_element
from xml.etree import ElementTree
import logging
import markdown
import re


logger = logging.getLogger(__name__)


class TypographyExtension(Extension):
    """
    Minor typographic improvements
    """
    def extendMarkdown(self, md):
        inline_processor = InlineProcessor(md)

        sectionPattern = SubstituteTextPattern(r'§ ', ('§&nbsp;',), md)
        inline_processor.inlinePatterns.register(sectionPattern, 'typo-section', 10)

        arrowPattern = SubstituteTextPattern(r' ➞', ('&nbsp;➞',), md)
        inline_processor.inlinePatterns.register(arrowPattern, 'typo-arrow', 10)

        ellipsisPattern = SubstituteTextPattern(r'\.\.\.', ('&hellip;',), md)
        inline_processor.inlinePatterns.register(ellipsisPattern, 'typo-ellipsis', 10)

        ellipsisPattern = SubstituteTextPattern(r' - ', ('&nbsp;–&nbsp;',), md)
        inline_processor.inlinePatterns.register(ellipsisPattern, 'typo-emdash', 10)

        squaredPattern = SubstituteTextPattern(r'\^2\^', ('²',), md)
        inline_processor.inlinePatterns.register(squaredPattern, 'squared', 65)

        cubedPattern = SubstituteTextPattern(r'\^3\^', ('³',), md)
        inline_processor.inlinePatterns.register(cubedPattern, 'cubed', 65)

        md.treeprocessors.register(inline_processor, 'typography', 2)


class JinjaPreprocessor(Preprocessor):
    """
    Ignore Jinja {{ ... }} and {% ... %} tags.
    """
    JINJA_RE = re.compile('({{([^}]+)}})|({%([^}]+)%})', re.MULTILINE | re.DOTALL)

    def run(self, lines):
        text = "\n".join(lines)

        def replace_match(match):
            return self.md.htmlStash.store(match[0])

        return re.sub(self.JINJA_RE, replace_match, text).split("\n")


class JinjaHtmlPostProcessor(RawHtmlPostprocessor):
    """
    Patch RawHtmlPostprocessor to recognize {% ... %} tags as block-level
    elements, and prevent them from being wrapped in a <p> tag.
    """
    JINJA_BLOCK_RE = re.compile('^{%([^}]+)%}$', re.MULTILINE | re.DOTALL)

    def isblocklevel(self, html):
        m = self.JINJA_BLOCK_RE.match(html)
        if m:
            return True
        return super().isblocklevel(html)


class JinjaExtension(Extension):
    """
    Escape Jinja {% ... %} and {{ ... }} statements in Markdown files.
    """
    def extendMarkdown(self, md):
        md.postprocessors.deregister('raw_html')
        md.preprocessors.register(JinjaPreprocessor(md), 'jinja', 25)
        md.postprocessors.register(JinjaHtmlPostProcessor(md), 'raw_html', 30)


class JinjaCurrencyPreprocessor(Preprocessor):
    """
    Wraps jinja template variables followed with "€" in a <span class="currency"> tag
    """
    JINJA_RE = re.compile('({{([^}]+)}})€', re.MULTILINE | re.DOTALL)

    def run(self, lines):
        text = "\n".join(lines)

        def replace_match(match):
            placeholder = self.md.htmlStash.store(f'<span class="currency">{match[1]}</span>€')
            return placeholder

        return re.sub(self.JINJA_RE, replace_match, text).split("\n")


class CurrencyExtension(Extension):
    """
    Wraps currency in a <span class="currency"> tag
    """
    def extendMarkdown(self, md):
        inline_processor = InlineProcessor(md)

        # 1,234.56€
        currencyPattern = SubstituteTextPattern(
            r'((\d+(,\d{3})*(\.\d{2})?))€',
            ('<span class="currency">', 1, '</span>€'), md
        )
        inline_processor.inlinePatterns.register(currencyPattern, 'currency', 65)
        md.treeprocessors.register(inline_processor, 'currency', 2)

        md.preprocessors.register(JinjaCurrencyPreprocessor(md), 'jinja-cur', 26)


class ResponsiveImageProcessor(Treeprocessor):
    allowed_parents = (
        'a', 'p', 'pre', 'ul', 'ol', 'dl', 'div', 'blockquote', 'noscript', 'section', 'nav', 'article',
        'aside', 'header', 'footer', 'table', 'form', 'fieldset', 'menu', 'canvas', 'details'
    )

    def __init__(self, md):
        super().__init__(md)

    def _swap_element(self, parent, old, new):
        """
        Replaces `old` element with `new` in `parent` element
        """
        for index, element in enumerate(parent):
            if element == old:
                parent[index] = new
                new.tail = old.tail
                return

    def _upgrade_img(self, img, parents):
        # Create <picture> with <source> for the different image types and sizes
        # Only apply to local images
        img_src = img.attrib.get('src')
        if img_src.startswith('/') or img_src.startswith(config.site_url + '/'):
            image_path = Path(img_src.removeprefix(config.site_url).removeprefix('/'))

            parent = parents[0]
            grandparent = parents[1]

            def has_single_child(element):
                return len(element) == 1 and not element.text

            img_attrs = img.attrib
            a_attrs = None
            image_maker = make_figure_element

            element_to_swap = img
            containing_element = parent

            # A valid <figure> parent with an empty <a> wrapping this <img>
            # In this case, wrap the <figure> around the <a>
            # li > a > img becomes li > figure > a > picture
            if (
                parent.tag == 'a'
                and has_single_child(parent)
                and grandparent.tag in self.allowed_parents
            ):
                a_attrs = parent.attrib
                element_to_swap = parent
                containing_element = grandparent

                # An empty <p> with an empty <a> with this <img>
                # Replace the whole thing with a <figure> containing the <a>
                # p > a > img becomes p > figure > a > picture
                if grandparent.tag == 'p' and has_single_child(grandparent):
                    element_to_swap = grandparent

                    greatgrandparent = parents[2]
                    containing_element = greatgrandparent

            # An empty <p> with this <img>
            # p > img becomes figure > picture
            elif parent.tag == 'p' and has_single_child(parent):
                element_to_swap = parent
                containing_element = grandparent

            # This element does not allow a <figure>. Just use a <picture>.
            elif parent.tag not in self.allowed_parents:
                image_maker = make_picture_element

            image = image_maker(image_path, config.output_path, img_attrs, a_attrs)

            self._swap_element(containing_element, element_to_swap, image)

    def run(self, root):
        parent_map = {}
        for parent in root.iter():
            for child in parent:
                parent_map[child] = parent

        for img in root.iter('img'):
            child = img
            parents = []
            while parent := parent_map.get(child):
                parents.append(parent)
                child = parent

            self._upgrade_img(img, parents)


class ResponsiveImagesExtension(Extension):
    """
    Transforms how <img> tags are rendered:

    - Adds srcset= to images to make them responsive
    - Wraps block images in a <figure> tag, and replaces the title with a <figcaption>

    """
    def extendMarkdown(self, md):
        md.registerExtension(self)
        self.md = md
        self.reset()
        md.treeprocessors.register(ResponsiveImageProcessor(md), 'figure', 0)

    def reset(self):
        pass


class SuperscriptExtension(Extension):
    """
    ^text^ is converted to <sup>text</sup>
    """

    # match ^, at least one character that is not ^, and ^ again
    SUPERSCRIPT_RE = r"(\^)([^\^]+)\2"

    def extendMarkdown(self, md):
        """Insert 'superscript' pattern before 'not_strong' pattern (priority 70)."""

        md.inlinePatterns.register(SimpleTagPattern(self.SUPERSCRIPT_RE, "sup"), 'superscript', 60)


class CustomFootnotesExtension(FootnoteExtension):
    def extendMarkdown(self, md):
        super().extendMarkdown(md)

    def makeFootnotesDiv(self, root):
        if not list(self.footnotes.keys()):
            return None

        container = ElementTree.Element("details")
        container.set("class", "footnote")
        container.set("id", "footnotes")
        summary = ElementTree.SubElement(container, "summary")
        summary.text = "Sources and footnotes"
        ol = ElementTree.SubElement(container, "ol")

        surrogate_parent = ElementTree.Element("div")

        backlink_title = self.getConfig("BACKLINK_TITLE")

        for index, id in enumerate(self.footnotes.keys(), start=1):
            li = ElementTree.SubElement(ol, "li")
            li.set("id", self.makeFootnoteId(id))
            # Parse footnote with surrogate parent as li cannot be used.
            # List block handlers have special logic to deal with li.
            # When we are done parsing, we will copy everything over to li.
            self.parser.parseChunk(surrogate_parent, self.footnotes[id])
            for el in list(surrogate_parent):
                li.append(el)
                surrogate_parent.remove(el)
            backlink = ElementTree.Element("a")
            backlink.set("href", f"#{self.makeFootnoteRefId(id)}")
            backlink.set("class", "footnote-backref")
            backlink.set(
                "title",
                backlink_title.format(index)
            )
            backlink.text = FN_BACKLINK_TEXT

            if len(li):
                node = li[-1]
                if node.tag == "p":
                    node.text = node.text + NBSP_PLACEHOLDER
                    node.append(backlink)
                else:
                    p = ElementTree.SubElement(li, "p")
                    p.append(backlink)

        return container


def patched_slugify(value, separator, keep_unicode=False):
    return slugify(value.lstrip(' 0123456789'), separator, keep_unicode)


class MarkdownProcessor(EntryContextProcessor):
    def __init__(self):
        super().__init__()

        self.markdown = markdown.Markdown(extensions=[
            'fenced_code',
            'meta',
            'tables',
            TocExtension(slugify=patched_slugify),
            CustomFootnotesExtension(BACKLINK_TEXT="⤴"),
            JinjaExtension(),
            SuperscriptExtension(),
            TypographyExtension(),
            CurrencyExtension(),
            ResponsiveImagesExtension(),
            WikiLinkExtension(
                base_url=config.wikilinks_base_url + '/',
                end_url=config.wikilinks_url_suffix,
                build_url=config.wikilinks_url_builder or build_url,
                html_class=config.wikilinks_html_class,
            ),
        ])

    def _parse_metadata(self, raw_metadata):
        metadata = {}
        for key, value in raw_metadata.items():
            if len(value) == 0:
                continue
            if len(value) == 1:
                value = value[0]

            if key.startswith('date_'):
                value = datetime.strptime(value, '%Y-%m-%d')

            if key.startswith('related_'):
                values = [v.strip() for v in value.split(',')]
                value = values if len(values) > 1 else values[0]

            metadata[key] = value
        return metadata

    def process_entry(self, entry_uri: str, entry_context: dict) -> dict:
        if entry_uri.endswith('.md'):
            with (config.content_path / entry_uri).open(encoding='utf-8') as f:
                html = self.markdown.reset().convert(f.read())

            entry_context.update({
                **self._parse_metadata(self.markdown.Meta),
                'body': html,
                'table_of_contents': self.markdown.toc_tokens,
                'url': f"{config.site_url}/{str(Path(entry_uri).with_suffix(config.html_url_extension))}",
            })

        return entry_context



from importlib import import_module
from pathlib import Path
from PIL import Image
from typing import Iterator
from ursus.config import config
from xml.etree import ElementTree
import fitz
import shutil
import sys


def import_class(import_path):
    module_name, class_name = import_path.rsplit('.', 1)
    module = import_module(module_name)
    return getattr(module, class_name)


def load_config(module_or_path: str) -> dict:
    """
    Imports the `config` variable from a Python file or a Python module

    Args:
        module_or_path (str): path.to.config.module, or path/to/config.py

    Returns:
        dict: The ursus config at this location
    """
    file_path = Path(module_or_path)
    if file_path.exists():
        sys.path.append(str(file_path.parent))
        module = import_module(file_path.with_suffix('').name)
    else:
        module = import_module(module_or_path)

    return getattr(module, 'config')


def is_ignored_file(path: Path, root_path: Path = None) -> bool:
    """Returns whether a file should be ignored by ursus.
    Args:
        path (Path): Path to the file
        root_path (Path, optional): Root path. The name of parent directories above root_path are not considered.

    Returns:
        bool: True if the file or any of its parents (up to root_path) start with _ or .
    """
    return (
        path.stem.startswith(('_', '.'))
        or any([
            p.stem.startswith(('_', '.'))
            for p in path.relative_to(root_path).parents
        ])
    )


def is_image(path: Path) -> bool:
    """Whether the given path points to an image

    Args:
        path (Path): The path to the file

    Returns:
        bool: Whether this file is an image
    """
    assert path.is_absolute(), 'is_image must be called with an absolute path'

    # Notably missing: .heif, .heic
    image_suffixes = ('.apng', '.avif', '.gif', '.jpg', '.jpeg', '.png', '.svg', '.webp')
    return path.is_file() and path.suffix.lower() in image_suffixes


def is_pdf(path: Path) -> bool:
    """Whether the given path points to a PDF

    Args:
        path (Path): The path to the file

    Returns:
        bool: Whether this file is a PDF
    """
    assert path.is_absolute(), 'is_pdf must be called with an absolute path'
    return path.is_file() and path.suffix.lower() == '.pdf'


def is_svg(path: Path):
    """Whether the given path points to an SVG image

    Args:
        path (Path): The path to the file

    Returns:
        bool: Whether this file is an SVG image
    """
    assert path.is_absolute(), 'is_svg must be called with an absolute path'
    return path.is_file() and path.suffix.lower() == '.svg'


def get_files_in_path(path: Path, whitelist: set = None, suffix: str = None) -> list[Path]:
    """
    Returns a list of valid, visible files under a given path.

    Args:
        path (Path): The path under which to find files
        whitelist (set, optional): Only include files that are part of this whitelist
        suffix (str, optional): Only include files with this suffix
    Returns:
        list[Path]: A list of files in this path
    """
    if whitelist:
        files = []
        for f in whitelist:
            if (not f.is_absolute()) and (path / f).exists():
                files.append(path / f)
            elif f.is_absolute() and f.is_relative_to(path):
                files.append(f)
    else:
        files = path.rglob('[!._]*' + (suffix or ''))

    return [
        f.relative_to(path) for f in files
        if f.is_file() and not is_ignored_file(f, path)
    ]


def copy_file(input_path: Path, output_path: Path):
    """Copies a file

    Args:
        input_path (Path): The absolute path of the file to copy
        output_path (Path): The absolute path of the file destination
    """
    assert input_path.is_absolute(), f"input_path {str(input_path)} is relative. It must be absolute."
    assert output_path.is_absolute(), f"output_path {str(output_path)} is relative. It must be absolute."

    output_path.parent.mkdir(parents=True, exist_ok=True)
    shutil.copy(input_path, output_path)


def make_image_thumbnail(pil_image: Image, max_size, output_path: Path):
    """Creates a thumbnail of an image. Strips EXIF metadata.

    Args:
        pil_image (Image): A Pillow Image object containing the image to resize
        max_size (TYPE): Max width and height of the preview image
        output_path (Path): Path to the resulting preview
    """
    assert output_path.is_absolute(), f"output_path {str(output_path)} is relative. It must be absolute."

    pil_image.thumbnail(max_size, Image.ANTIALIAS)
    save_args = {'optimize': True}
    if output_path.suffix.lower() == '.jpg':
        save_args['progressive'] = True
    elif output_path.suffix.lower() == '.webp':
        save_args['exact'] = True

    output_path.parent.mkdir(parents=True, exist_ok=True)

    # Note: The saved image is stripped of EXIF data
    pil_image.save(output_path, **save_args)


def make_pdf_thumbnail(pdf_path: Path, max_size, output_path: Path):
    """Creates an image preview of a PDF file

    Args:
        pdf_path (Path): Path to the PDF file to preview
        max_size (TYPE): Max width and height of the preview image
        output_path (Path): Path to the resulting preview
    """
    assert output_path.is_absolute(), f"output_path {str(output_path)} is relative. It must be absolute."

    width, height = max_size
    doc = fitz.open(pdf_path)
    pixmap = doc[0].get_pixmap(alpha=False)
    thumbnail = Image.frombytes('RGB', [pixmap.width, pixmap.height], pixmap.samples)
    make_image_thumbnail(thumbnail, max_size, output_path)


def get_image_transforms(original_path: Path) -> Iterator[dict]:
    """
    Yields a list of image transforms that apply to a file.

    Args:
        original_path (Path): Path of the original file/image to transform

    Yields:
        Iterator[dict]: A list of image transforms that apply to this file.
    """
    suffix_to_mimetype = {
        '.apng': 'image/apng',
        '.avif': 'image/avif',
        '.gif': 'image/avif',
        '.jpg': 'image/jpeg',
        '.jpeg': 'image/jpeg',
        '.pdf': 'application/pdf',
        '.png': 'image/png',
        '.svg': 'image/svg+xml',
        '.webp': 'image/webp',
    }

    for key, transform in config.image_transforms.items():
        includes = transform.get('include', ['*'])
        includes = [includes] if isinstance(includes, str) else includes

        excludes = transform.get('exclude', [])
        excludes = [excludes] if isinstance(excludes, str) else excludes

        transform_applies_to_file = (
            any(original_path.match(pattern) for pattern in includes)
            and not any(original_path.match(pattern) for pattern in excludes)
        )

        if transform_applies_to_file:
            # Normalise and deduplicate suffixes
            # For orig_image.JPG, ('original', 'jpg') becomes ('.jpg')
            output_suffixes = set([
                original_path.suffix.lower() if t == 'original' else '.' + t
                for t in transform.get('output_types', ['original'])
            ])

            for suffix in output_suffixes:
                if suffix == original_path.suffix.lower():
                    output_image_path = original_path
                else:
                    if original_path.suffix.lower() == '.svg':
                        # .svg images are not be converted
                        continue

                    output_image_path = original_path.with_suffix(suffix)

                yield {
                    **transform,
                    'is_default': key == '',
                    'input_mimetype': suffix_to_mimetype[original_path.suffix.lower()],
                    'output_mimetype': suffix_to_mimetype[output_image_path.suffix.lower()],
                    'output_path': output_image_path.parent / key / output_image_path.name,  # It works if key is empty
                }


def make_picture_element(original_path: Path, output_path: Path, img_attrs={}):
    """
    Creates a responsive HTML <picture> element
    """
    # Mimetypes with broad support that don't require their own <source> element
    standard_mimetypes = ('image/png', 'image/jpeg', 'image/svg+xml')

    # TODO: SVG? PDF?
    default_src = None

    # Build a list of srcsets grouped by mimetype
    sources_by_mimetype = {}
    for transform in get_image_transforms(original_path):
        width = transform['max_size'][0]
        mimetype = transform['output_mimetype']

        if mimetype.startswith('image/'):
            srcset_part = f"{config.site_url}/{str(transform['output_path'])} {width}w"
            sources_by_mimetype.setdefault(mimetype, [])
            sources_by_mimetype[mimetype].append(srcset_part)

            if not default_src and mimetype in standard_mimetypes:
                default_src = transform['output_path']

    # Create a <picture> with <source type="" srcset=""> for each mimetype
    picture = ElementTree.Element('picture')
    for mimetype, srcset_elements in sources_by_mimetype.items():
        source = ElementTree.Element('source', attrib={
            'type': mimetype,
            'srcset': ", ".join(srcset_elements)
        })
        picture.append(source)

    # Add an <img> with the default image to the <picture>
    img = ElementTree.Element('img', attrib=img_attrs)
    assert default_src is not None, f"default_src is None for {original_path}"

    # The output image might not exist yet, so we use the source's dimensions
    abs_original_path = config.content_path / original_path
    if abs_original_path.suffix not in ('.svg', '.pdf'):
        with Image.open(abs_original_path) as pil_image:
            width, height = pil_image.size
            img.attrib['width'] = str(width)
            img.attrib['height'] = str(height)
    img.attrib['loading'] = 'lazy'
    img.attrib['src'] = f"{config.site_url}/{str(default_src)}"

    picture.append(img)

    return picture


def make_figure_element(original_path: Path, output_path: Path, img_attrs={}, a_attrs=None):
    """
    Creates a responsive HTML <figure> element with the image title as <figcaption>. Returns a simple <picture> if there
    is no title.
    """
    image = make_picture_element(original_path, output_path, img_attrs)
    if a_attrs and a_attrs.get('href'):
        a_attrs['target'] = '_blank'
        wrapped_image = ElementTree.Element('a', a_attrs)
        wrapped_image.append(image)
    else:
        wrapped_image = image

    if not img_attrs.get('title'):
        return wrapped_image

    figure = ElementTree.Element('figure')
    figure.append(wrapped_image)
    figcaption = ElementTree.Element('figcaption')
    figcaption.text = img_attrs['title']
    figure.append(figcaption)

    return figure



from . import Renderer
from lunr import lunr
from pathlib import Path
from ursus.config import config
import json
import logging


logger = logging.getLogger(__name__)


class LunrIndexRenderer(Renderer):
    """
    Renders a .json search index for Lunr.js. The resulting file is a dictionary
    with two keys:

    - 'index': An index object that can be passed to Lunr.js
    - 'documents': A dict of entry URI to documents that can be used to render
        search results (titles, URLs, excerpts, etc.)
    """
    def __init__(self):
        super().__init__()

    def render(self, context: dict, changed_files: set = None) -> set:
        index_output_path = config.output_path / config.lunr_index_output_path
        logger.info(f"Generating search index at {index_output_path}")

        indexed_documents = []
        returned_documents = {}

        for index in config.lunr_indexes.get('indexes', []):
            for entry_uri, entry in context['entries'].items():
                if Path(entry_uri).match(index['uri_pattern']):
                    indexed_documents.append((
                        {
                            'uri': entry_uri,
                            **{field: entry.get(field, '') for field in config.lunr_indexes['indexed_fields']},
                        },
                        {
                            'boost': index.get('boost', 1)
                        }
                    ))

                    returned_documents[entry_uri] = {
                        field: entry.get(field) for field in index['returned_fields']
                    }

        index = lunr(
            ref='uri',
            fields=config.lunr_indexes.get('indexed_fields', []),
            documents=indexed_documents
        )

        index_output_path.parent.mkdir(parents=True, exist_ok=True)
        with index_output_path.open('w+') as index_file:
            json.dump({
                'index': index.serialize(),
                'documents': returned_documents,
            }, index_file)

        return set([config.lunr_index_output_path])
